from laundromat.regex_engine.regex_base import RegexBase
import re


class RegexAmount(RegexBase):
    """
    Amount is percent
    """
    @property
    def regex_pattern(self):
        return r"\d{1,3}\s*(%|prosent)"


    @property
    def context(self):
        return ["amount"]

    @property
    def label(self):
        return "AMOUNT"

    @property
    def score(self):
        return 1

    def validate(self):
        pass



from laundromat.regex_engine.regex_base import RegexBase
import re

class RegexAccountNumber(RegexBase):
    @property
    def regex_pattern(self):
        """
        TODO Add validation or change pattern as this one overlaps with personnummer
        """
        return r"(?<!\d)(\d{11})|(\d{4}(\.|\s)\d{2}(\.|\s)\d{5}\b)(?!\s*\d)"

    @property
    def context(self):
        return ["Tid og/eller dato"]

    @property
    def label(self):
        return "DTM"

    @property
    def score(self):
        return 1

    def validate(self):
        pass

    def validate_acc(self, acc: str):
        #TODO add a step that puts string in the right format
        # so as to avoid xx.xx.xxxxx for example
        d1 = int(acc[0])
        d2 = int(acc[1])
        d3 = int(acc[2])
        d4 = int(acc[3])
        d5 = int(acc[4])
        d6 = int(acc[5])
        d7 = int(acc[6])
        d8 = int(acc[7])
        d9 = int(acc[8])
        d10 = int(acc[9])

        control_digit = int(acc[10])

        mod_11 = (2*d10+3*d9+4*d8+5*d7+6*d6+7*d5+2*d4+3*d3+4*d2+5*d1)%11
        if mod_11+control_digit ==11:
            return 1.0
        else:
            return 0.0



from laundromat.regex_engine.regex_base import RegexBase
import spacy
import re
import numpy as np


def levenstein_distance(pnr):
    """
    Function that checks the Levenstein distance of a string from a pnr
    :param pnr: The inputted pnr
    :return: Levenstein distance of from a PNR
    """
    # TODO: Make this to check for distance to valid PNR, so far it only checks if it is 11 digits

    regex_size = 11
    size_x = len(pnr) + 1
    size_y = regex_size + 1  # size of pnr
    matrix = np.zeros((size_x, size_y))  # Empty matrix

    for x in range(size_x):
        matrix[x, 0] = x
    for y in range(size_y):
        matrix[0, y] = y

    for x in range(1, size_x):
        for y in range(1, size_y):
            if re.match(r"\d", pnr[x - 1]):
                matrix[x, y] = min(
                    matrix[x - 1, y] + 1,
                    matrix[x - 1, y - 1],
                    matrix[x, y - 1] + 1
                )
            else:
                matrix[x, y] = min(
                    matrix[x - 1, y] + 1,
                    matrix[x - 1, y - 1] + 1,
                    matrix[x, y - 1] + 1
                )

    print('The Levenstein distance from a pnr is {0}'.format(matrix[size_x - 1, size_y - 1]))
    return matrix[size_x - 1, size_y - 1]


class RegexFnr(RegexBase):

    @property
    def regex_pattern(self):
        return r"((?<!\d)(\d{11}\b)|(\b\d{6}\s\d{5}))(\.*)(?!\s*\d)"

    @property
    def context(self):
        return ["fødselnummer"]

    @property
    def label(self):
        return "FNR"

    @property
    def score(self):
        return 1

    @property
    def validate(self):
        pass

    def validate_pnr(self, pnr: str):
        """
        :param pnr:
        :return:
        """

        if len(pnr) == 11 and re.match(r"\d{11}", pnr):
            d1 = int(pnr[0])
            d2 = int(pnr[1])

            m1 = int(pnr[2])
            m2 = int(pnr[3])

            y1 = int(pnr[4])
            y2 = int(pnr[5])

            i1 = int(pnr[6])
            i2 = int(pnr[7])
            i3 = int(pnr[8])

            k1 = int(pnr[9])
            k2 = int(pnr[10])

            k1_mod = (3 * d1 + 7 * d2 + 6 * m1 + m2 + 8 * y1 + 9 * y2 + 4 * i1 + 5 * i2 + 2 * i3) % 11
            new_k1 = 0 if k1_mod == 0 else (11 - k1_mod)

            k2_mod = (5 * d1 + 4 * d2 + 3 * m1 + 2 * m2 + 7 * y1 + 6 * y2 + 5 * i1 + 4 * i2 + 3 * i3 + 2 * k1) % 11
            new_k2 = 0 if k2_mod == 0 else (11 - k2_mod)

            if k1 != new_k1 or k2 != new_k2:
                print('not a valid PNR')
                levenstein_distance(pnr)

            if k1 == new_k1 and k2 == new_k2:
                print('Valid PNR')

                return 1.0
              
        else:
            levenstein_distance(pnr)


from spacy.tokens import Doc

def merger(doc):
    """
    Merges the RegEx entities and the NER entities. If conflicts arise the function will try to create the
    largest entity possible. Whether by merging the entities or choosing the larger of the two.
    """
    list_ner, list_regex = doc.ents, doc._.ents_regex

    #Removes duplicates from regex
    correct_regex = []
    start_index_list = []
    end_index_list = []
    for ent in list_regex:
        if ent.start in start_index_list or ent.end in end_index_list:
            continue
        else:
            correct_regex.append(ent)
            start_index_list.append(ent.start_char)
            end_index_list.append(ent.end_char)
    
    #Merges the two lists
    merged = list(list_ner)
    start_index_list = [ent.start for ent in list_ner]
    end_index_list = [ent.end for ent in list_ner]
    for ent in correct_regex:
        if ent.start in start_index_list or ent.end in end_index_list:
            continue
        else:
            merged.append(ent)
            start_index_list.append(ent.start)
            end_index_list.append(ent.end)
    merged.sort(key=lambda x: x.start)

    #Resolves overlaps
    final = merged.copy()
    if merged:
        for i in range(len(merged)):
            if(i==len(merged)-1):
                continue
            start_a = merged[i].start_char
            start_b = merged[i+1].start_char
            end_a = merged[i].end_char
            end_b = merged[i+1].end_char

            if start_a == start_b or end_a==end_b:
                start = min(start_a, start_b)
                end = max(end_a, end_b)
                del final[i:i+2]
                new_ent = doc.char_span(start, end, merged[i].label)
                final.append(new_ent)
            elif end_a > start_b:
                start = min(start_a, start_b)
                end = max(end_a, end_b)
                del final[i:i+2]
                new_ent = doc.char_span(start, end, merged[i].label)
                final.append(new_ent)

    doc.ents = final
    return doc


from laundromat.regex_engine.regex_base import RegexBase


class RegexCreditCard(RegexBase):

    @property
    def regex_pattern(self):
        return r"\b((4\d{3})|(5[0-5]\d{2})|(6\d{3})|(1\d{3})|(3\d{3}))[- ]?(\d{3,4})[- ]?(\d{3,4})[- ]?(\d{3,5})\b"

    @property
    def context(self):
        return [
            "credit",
            "card",
            "visa",
            "mastercard",
            "cc ",
            "amex",
            "discover",
            "jcb",
            "diners",
            "maestro",
            "instapayment"
        ]

    @property
    def label(self):
        return "CREDIT_CARD"

    @property
    def score(self):
        return 1

    def validate(self):
        pass



from laundromat.regex_engine.regex_base import RegexBase
import datetime



class RegexGeneric(RegexBase):

    def __init__(self, pattern = None, context = None, label = None):
        if not pattern:
            raise ValueError("Your RegEx must have a value")
        if not context:
            raise ValueError("Your RegEx must have a context")
        if not label:
            raise ValueError("Your RegEx must have a label")
        
        self.__regex_pattern = pattern
        self.__context = context
        self.__label = label

    
    @property
    def regex_pattern(self):
        return self.__regex_pattern
            

    @property
    def context(self):
        return self.__context

    @property
    def label(self):
        return self.__label

    @property
    def score(self):
        return 1

    def validate(self):
        pass


import re

from abc import ABC, abstractmethod
from spacy.tokens import Doc


class RegexBase(ABC):
    """
    Abstract class that RegEx classes inherit from.
    """
    @property
    @abstractmethod
    def regex_pattern(self):
        """
        :return: Returns the regex pattern of the class
        """
        return NotImplementedError()

    @property
    @abstractmethod
    def label(self):
        """
        :return: Return label to substitute with
        """
        return NotImplementedError()

    @property
    @abstractmethod
    def score(self):
        return NotImplementedError()

    @property
    @abstractmethod
    def context(self):
        """
        :return: List of pattern groups
        """
        return NotImplementedError()

    @abstractmethod
    def validate(self):
        return NotImplementedError()

    def sub_all(self, text: str):
        return re.sub(self.regex_pattern, f'[{self.label}]', text)

    def find_all(self, text: str):
        hits = []
        for hit in re.finditer(self.regex_pattern, text):
            hits.append({
                'start': hit.start(),
                'end': hit.end(),
                'hit': hit.group(),
                'label': self.label
                         })

        return hits

    def spacy_pipe_func(self, doc: Doc) -> Doc:
        """
        :param doc: Spacy doc
        :return: Spacy doc
        """
        pass



from datetime import datetime
from itertools import zip_longest
from pathlib import Path
import warnings
import random
import networkx as nx
import pandas as pd
import spacy
from laundromat.spacy.list_merger import merger
from laundromat.spacy.matcher_list import ListMatcher
from laundromat.spacy.matcher_regex import RegexMatcher
import laundromat.spacy.matcher_regex as mreg
from laundromat.regex_engine.generic import RegexGeneric
from sklearn.metrics import f1_score
from spacy import displacy
from spacy.gold import GoldParse
from spacy.matcher import Matcher
from spacy.scorer import Scorer
from spacy.tokens import Doc
from spacy.util import compounding, minibatch


class SpacyModel:
    """
    :SpacyModel class: A class for managing a SpaCy nlp model with methods for the following:
    - adding custom RegEx 
    - easy training and display of results
    - various metrics
    """
    
    def __init__(self, model_path=None):
        if not model_path:
            self.model = spacy.load("nb_core_news_lg")
        else:
            self.model = spacy.load(model_path)
            self.model_path = model_path
        self.matcher = Matcher(self.model.vocab)
        Doc.set_extension('ents_regex', force=True, default=True)
        self.patterns_added = False
        self.list_matcher = ListMatcher()
        self.regex_matcher = RegexMatcher()

    def add_patterns(self, entities: list = None, before_ner = False, lookup = False):
        """
        Adds desired patterns to the entity ruler of the SpaCy model.

        :param entities: a list of strings denoting which entities the nlp model should detect.
        """
        if self.patterns_added:
            self.model = spacy.load(self.model_path)
            self.patterns_added = False
        ruler = self.list_matcher.csv_list_matcher(nlp=self.model)
        if before_ner:
            self.model.add_pipe(self.regex_matcher.match_func, name="regex_matcher", before='ner')
        else:
            self.model.add_pipe(self.regex_matcher.match_func, name="regex_matcher", after='ner')
        if lookup:
            self.model.add_pipe(ruler, after="ner")
        self.model.add_pipe(merger, last = True)
        
    def new_regex(self, pattern: str, context: str, label: str):
        self.regex_matcher.append_regexes(RegexGeneric(pattern, context, label))

    def print_regex_labels(self):
        "Prints regex labels"
        for regex in self.regex_matcher.regexes:
            print(regex.label)

    def add_list(self, new_list):
        """
        :param new_list: new lists must be a tuple of a path and a tag.
        """
        self.list_matcher.default_list_append(new_list)

    def predict(self, text: str):
        """
        :param text: a string of text which is to be analysed.
        :return: found entities, their labels, start, and end index.
        """
        doc = self.model(text)
        ents = [[ent.text, ent.label_, ent.start, ent.end] for ent in doc.ents]
        return ents
        
    def predict_list(self, text_list):
        docs = list(self.model.pipe(text_list))
        ents = []
        for d in docs:
            ents.append([[ent.text, ent.label_, ent.start, ent.end] for ent in d.ents])
        return ents

    def doc(self, text: str):
        """
        Returns the doc, a SpaCy object containing metadata about the text.

        :param text: A text string to be ran through the model
        :return: A text in doc-format
        """
        return self.model(text)

    def doc_list(self, text_list):
        return list(self.model.pipe(text_list))

    def display(self, text: str):
        colors = {"PER": "#FFA3A3",
                  "FNR": "#FDB9A2",
                  "DTM": "#FBCEA0",
                  "TLF": "#D8DEAD",
                  "AMOUNT": "#B5EEB9",
                  "LOC": "#ADDFEE",
                  "CREDIT_CARD": "#C0C7EC",
                  "ORG": "D2AFEA"
                  }

        options = {"ents": ["ORG", "LOC", "PER", "FNR", "AMOUNT", "TLF", "DTM", "CREDIT_CARD"],
                   "colors": colors
                   }
        displacy.render(self.doc(text), style='ent', jupyter=True, options=options)

    def disable_ner(self):
        """
        Disables the NER-model in the pipeline.
        """
        self.disabled = self.model.disable_pipes("ner")

    def enable_ner(self):
        """
        Enables the NER-model in the pipeline.
        """
        self.disabled.restore()

    def pipeline(self):
        """
        Prints current NLP pipeline.
        """
        print(self.model.pipe_names)

    def replace(self, text: str, replacement = "entity", replacement_char = "~"):
        """
        Replaces found entities in the given text with the attendant entity labels,
        e.g. a name is replaced with <PER>. Shuffle replacement is hard coded to use a list
        of girl, boy, county, country, and town names.

        :param text: The text in question as a string.
        :param replacement: The kind of replacement desired. Default is entity.
        :param replacement_char: The character used for character and padding replacement. Default is ~.
        :returns: the modified text as a string
        """

        doc = self.model(text)
        censored_text = text
        ents = [[ent.text, ent.label_, ent.start, ent.end, "NA"] for ent in doc.ents]

        if replacement=="entity":
            for ent in ents:
                censored_text = censored_text.replace(ent[0], "<" + ent[1] + ">")
        elif replacement=="character":
            for ent in ents:
                censored_text = censored_text.replace(ent[0], replacement_char)
        elif replacement=="pad":
            for ent in ents:
                censored_text = censored_text.replace(ent[0], replacement_char*len(ent[0]))
        elif replacement=="shuffle":
            girls_names = self.list_matcher.get_data('jentefornavn_ssb.csv')['text']
            boys_names = self.list_matcher.get_data('guttefornavn_ssb.csv')['text']
            name_list = girls_names.append(boys_names, ignore_index=True)

            #kom_names = self.list_matcher.get_data('kommuner.csv')['text']
            country_names = self.list_matcher.get_data('land.csv')['text']
            #villages_names = self.list_matcher.get_data('tettsteder.csv')['text']
            #loc_list = kom_names.append(country_names.append(villages_names, ignore_index=True), ignore_index=True)
            loc_list = country_names
            for ent in ents:
                if ent[1] == 'PER':
                    censored_text = censored_text.replace(ent[0], name_list[np.random.randint(0, len(name_list))])
                if ent[1] == 'LOC':
                    censored_text = censored_text.replace(ent[0], loc_list[np.random.randint(0, len(loc_list))])
                else:
                    censored_text = censored_text.replace(ent[0], "<" + ent[1] + ">")
        else:
            raise ValueError("replacement must be either entity, character, pad, or shuffle")
        return censored_text

    def train(self, TRAIN_DATA, labels: list, n_iter: int = 30, output_dir=None):
        """
        :param TRAIN_DATA: training data
        :param labels: texts with labels
        :param n_iter: number
        :param output_dir: Where the model should be saved
        """

        ner = self.model.get_pipe("ner")
        for lab in labels:
            ner.add_label(lab)
        optimizer = self.model.resume_training()
        move_names = list(ner.move_names)
        pipe_exceptions = ["ner", "trf_wordpiecer", "trf_tok2vec"]
        other_pipes = [pipe for pipe in self.model.pipe_names if pipe not in pipe_exceptions]
        with (self.model.disable_pipes(*other_pipes)), warnings.catch_warnings():
            warnings.filterwarnings("once", category=UserWarning, module='spacy')
            sizes = compounding(1.0, 4.0, 1.001)
            # batch up the examples using spaCy's minibatch
            for itn in range(n_iter):
                random.shuffle(TRAIN_DATA)
                batches = minibatch(TRAIN_DATA, size=sizes)
                losses = {}
                for batch in batches:
                    texts, annotations = zip(*batch)
                    self.model.update(texts, annotations, sgd=optimizer, drop=0.35, losses=losses)
                print("Losses", losses)
        # Save model
        if output_dir is not None:
            output_dir = Path(output_dir)
            if not output_dir.exists():
                output_dir.mkdir()
            current_dtm = str(datetime.now)
            alphanumeric_dtm = [character for character in current_dtm if character.isalnum()]
            alphanumeric_dtm = "".join(alphanumeric_dtm)
            self.model.meta["name"] = "ner_model_"+alphanumeric_dtm
            self.model.to_disk(output_dir)
            print("Saved model to", output_dir)

            # test the saved model
            print("Loading from", output_dir)
            nlp2 = spacy.load(output_dir)
            # Check the classes have loaded back consistently
            assert nlp2.get_pipe("ner").move_names == move_names

    def f1_scorer(self, TEST_DATA):
        """
        Ignores data of the wrong format. If this is a problem, use print_scores instead.
        :return: Scores as a dictionary.
        """
        scorer = Scorer()
        df = pd.DataFrame(TEST_DATA)
        df.columns = ["Text", "True_entities"]
        for txt, ents in zip(df["Text"], df["True_entities"]):
            doc = self.model.make_doc(txt)

            #Gold refers to the correct entity labels
            gold = GoldParse(doc, entities=ents["entities"])
            pred = self.model(txt)
            scorer.score(pred, gold)
        return scorer.scores

    def confusion_matrix(self, TEST_DATA, strict = True):
        """
        Calculates confusion matrix for given data. 
        Only considers whether a token has been labeled and not if it has been labeled correctly.

        :return: Confusion matrix as a dataframe.
        """
        tp, fn, fp, tn = 0, 0, 0, 0
        df = pd.DataFrame(TEST_DATA)
        df.columns = ["Text", "True_entities"]
        df["Model_entities"] = df["Text"].apply(lambda x: {"entities": [(ent.start_char, ent.end_char, ent.label_) for ent in self.model(x).ents]})
        df["total"] = df["Text"].apply(lambda x: len(self.model(x)))
        for model, truth in zip_longest(df["Model_entities"], df["True_entities"]):
            ents_m = model["entities"]
            ents_t = truth["entities"]
            overlap = 0
            for m in model["entities"]:
                for t in truth["entities"]:
                    if strict:
                        if (t[0] == m[0]) and (m[1] == t[1]):
                            tp += 1
                            overlap += 1
                    else:
                        if (t[0] <= m[0] <=t[1]) or (t[0] <= m[1] <= t[1]):
                            tp += 1
                            overlap += 1
            fp += len(ents_m) - overlap
            fn += len(ents_t) - overlap
            tn = tn -(len(ents_m)-overlap) -len(ents_t)
        tn += df["total"].sum()
        confusion = pd.DataFrame([[tp, fp], [fn, tn]], index = ["Predicted Positive", "Predicted Negative"], columns=["Is Positive", "Is Negative"])
        return confusion
    
    def print_scores(self, TEST_DATA, strict=True):
        """
        Prints the following scores:
        - Accuracy
        - Balanced accuracy
        - Precision
        - Recall
        - F1 score
        """
        cf = self.confusion_matrix(TEST_DATA, strict)
        true_positive = cf["Is Positive"].iloc[0]
        true_negative = cf["Is Negative"].iloc[1]
        false_positive = cf["Is Negative"].iloc[0]
        false_negative = cf["Is Positive"].iloc[1]
        accuracy = (true_positive+true_negative)/(cf.values.sum() + 1e-100)
        precision = true_positive/(true_positive+false_positive+ 1e-100)
        specificity = true_negative/(true_negative+false_positive+ 1e-100)
        recall = true_positive/(true_positive+false_negative+ 1e-100)
        balanced_accuracy = (recall+specificity)/2
        f_1 = 2*true_positive/(2*true_positive+false_positive+false_negative + 1e-100)

        print("Accuracy is: ", accuracy)
        print("Balanced accuracy is: ", balanced_accuracy)
        print("Precision is: ", precision)
        print("Recall is: ", recall)
        print("F1 score is: ", f_1)
    
    def count(self, text:str):
        ent_counts = {"PER": 0,
                  "FNR": 0,
                  "DTM": 0,
                  "TLF": 0,
                  "AMOUNT": 0,
                  "LOC": 0,
                  "CREDIT_CARD": 0,
                  "ORG": 0
                  }
        try:
            tx = str(text)
        except TypeError:
            return ent_counts
        for ent in self.model(tx).ents:
            try:
                ent_counts[ent.label_] = ent_counts[ent.label_] + 1
            except KeyError:
                ent_counts[ent.label_] = 1
        return ent_counts

    def dependency_graph(self, text: str):
        """
        :param text: The text in question as a string.
        :return: A digraph of the linguistic dependencies in the text. 
        """
        #TODO Seems to struggle with the  <ENTITY> format as <, >, and . end up as their own nodes.
        doc = self.model(text)
        edges = []
        for token in doc:
            for child in token.children:
                edges.append(('{0}'.format(token.lower_),
                            '{0}'.format(child.lower_)))
        return nx.DiGraph(edges)

    def top_n_nodes(self, text:str, n=10):
        """
        :param text: The text in question as a string.
        :param n: Number of nodes desired. Default is 10.
        :return: The top n nodes with highest degree in the dependency graph.
        """
        graph = self.dependency_graph(text)
        sorted_node_degrees = sorted(list(graph.degree), key= lambda x: x[1], reverse=True)
        if(n>len(sorted_node_degrees)):
            n = len(sorted_node_degrees)
        return sorted_node_degrees[:n]

    def similarity(self, text:str, replacement = "entity", replacement_char = "~"):
        """
        :param text: The text in question as a string.
        :param replacement: The kind of replacement desired. Default is entity.
        :param replacement_char: The character used for character and padding replacement. Default is ~.
        :return: The cosine similarity between the unaltered text and the altered text.
        """
        original = self.model(text)
        censored = self.model(self.replace(text, replacement,replacement_char))
        return original.similarity(censored)



from laundromat.regex_engine.regex_base import RegexBase
import re

class RegexTlfNr(RegexBase):
    @property
    def regex_pattern(self):
        """
        Searches for 8-digit phone numbers and country codes.
        """
        # Det regexuttrykket som retuneres er rimelig robust, men tar bare norske tlfnr
        return r"(?<!\d)((\+|00)47\s*)*(\d{8}|\d{3}\s\d{2}\s\d{3}|\d{2}\s\d{2}\s\d{2}\s\d{2})(\.*)(?!\s*\d)"

    @property
    def context(self):
        return ["Telefonnummer"]

    @property
    def label(self):
        return "TLF"

    @property
    def score(self):
        return 1

    def validate(self):
        pass

    @staticmethod
    def validate_tlf(tlf: str):

        tlf = tlf.replace(' ', '')
        tlf = re.sub(r"(\+|00)47", '', tlf)
        print(tlf)

        if int(tlf) in range(20000000, 80000000) or int(tlf) in range(90000000, 100000000):
            print('gyldig norsk nummer')
            return 1.0
        else:
            print('ugyldig nummer')


# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')



setup(
    name='laundromat',
    version='1.1.1',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.6',
    install_requires=install_requires,
    author="Marius Dioli @ NAV IKT and Petter Sunde Nymark @ NAV IKT",
    description="Scrub personal information from text with Laundromat.",
    project_description="",
    license="MIT",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    package_data={'laundromat': ['spacy/data/*.csv']}

)



def regex_formatter(entities: list = None):
    """
    Formats desired entities such that they can be fed to SpaCy.

    :param entities: A list of strings denoting which entities one wishes to include in the model.
    """
    labels = all_possible_labels()

    if not entities:
        regex = [ent.regex_pattern for ent in regex_engines()]
        form = []
        for label, reg in zip(labels, regex):
            form += [{"label": label, "pattern": [{"TEXT": {"REGEX": reg}}]}]
        return form
    elif set(entities).issubset(set(labels)):
        regex = [ent.regex_pattern for ent in regex_engines() if ent.label in entities]
        form = []
        for label, reg in zip(labels, regex):
            form += [{"label": label, "pattern": [{"TEXT": {"REGEX": reg}}]}]
        return form
    # TODO Else if they add entities that do not exist yet.





from laundromat.regex_engine.regex_base import RegexBase
import re

class RegexAccountNumber(RegexBase):
    @property
    def regex_pattern(self):
        """
        """
        return r"(\b(?<!(\d{1,11}))(\d{2,3})(?!\d{1,11})(?!\s\d{1,4})(?!\.\d{1,4})(?!\-\d{1,4})(?!\%)(?!\sprosent)\b)"

    @property
    def context(self):
        return ["Alder"]

    @property
    def label(self):
        return "AGE"

    @property
    def score(self):
        return 1

    def validate(self):
        pass

    def validate_age(self, age: str):
        if(int(age)<0 or int(age)>110):
            return 0.0
        else:
            return 0.5



import pandas as pd
import os
from spacy.pipeline import EntityRuler

class ListMatcher:

    def __init__(self):
        self.default_list = [('land.csv', "LOC"),
                            ('etternavn_ssb.csv', "PER"),
                            ('guttefornavn_ssb.csv', "PER"),
                            ('jentefornavn_ssb.csv', "PER")]

    def csv_list_matcher(self, nlp, path_list = None):
        """
        Reads csv-files for names and surnames and uses EntityRuler to match to a entity and return to be added to
        the model pipeline

        :param nlp: The Spacy model
        :param path_list: a list of tuples with the path of the list and its attendant entity.
        Accepts only single entity type lists. Lists must be a single column with the string "text" in the first cell.
        :return: The ruler containing the new pattern rules to be included in the pipeline
        """
        if path_list is None:
            path_list = self.default_list
        ruler = EntityRuler(nlp)
        print(path_list)
        for path, label in path_list:
            df = self.get_data(path)
            name_patterns = [{"label": label, "pattern": [{"lower": name.lower()}]} for name in df["text"]]
            ruler.add_patterns(name_patterns)
        return ruler

    def get_data(self, file_name:str):
        """
        Takes the file path of the integrated data and returns the csv-file as a dataframe
        :param file_name:
        :return dataframe:
        """
        location = os.path.dirname(os.path.realpath(__file__))
        my_file = os.path.join(location, 'data', file_name)
        dataframe = pd.read_csv(my_file, sep=';', encoding = 'utf-8')

        return dataframe
    
    def default_list_append(self, item):
        self.default_list.append(item)


from laundromat.regex_engine.regex_base import RegexBase
import datetime



class RegexDateTime(RegexBase):

    @property
    def regex_pattern(self):
        """
        Searches for dates and times in a variety of format. Does not currently support the day-month format
        (e.g. 12. July)
        TODO Currently overlaps with phone numbers. Maybe validation is the best way to get around this.
        TODO Add day and month names for a lookup
        """
        return r"(?<!\d)((\d{1,2}(\-|\\|\/|\.|\s)?(0|1)?(\d)(\-|\\|\/|\.|\s)?(19|20)?\d{2}))(\.*)(?!\s*\d)"

    @property
    def context(self):
        return ["Tid og/eller dato"]

    @property
    def label(self):
        return "DTM"

    @property
    def score(self):
        return 1

    def validate(self):
        pass

    @staticmethod
    def validate_dmt(dmt: str):
        """
        Checks a list of different datetime formats an does a checksum
        :param dmt:
        :return:
        """
        check_pass = False
        date_format = ['%d.%m.%y', '%d.%m.%Y',
                       '%d-%m-%y', '%d-%m-%Y',
                       '%d%m%y', '%d%m%Y',
                       '%d %m %y', '%d %m %Y']

        for form in date_format:
            try:
                datetime.datetime.strptime(dmt, form)
                check_pass = True
            except ValueError:
                pass

        if check_pass:
            return 1.0
        else:
            return 0



import re
from spacy.tokens import Doc, Span, Token
from laundromat.regex_engine.fnr import RegexFnr
from laundromat.regex_engine.credit_card import RegexCreditCard
from laundromat.regex_engine.tlfnr import RegexTlfNr
from laundromat.regex_engine.amount import RegexAmount
from laundromat.regex_engine.date_time import RegexDateTime
from laundromat.regex_engine.generic import RegexGeneric

class RegexMatcher:
    def __init__(self):
        self.regexes = [RegexFnr(),
                        RegexCreditCard(),
                        RegexTlfNr(),
                        RegexDateTime(),
                        RegexAmount()
                        ]
    

    def append_regexes(self, regex):
        self.regexes.append(regex)


    def match_func(self, doc):
        """
        Uses regex to find patterns and label them with the given label.

        :param doc: The text in doc-format
        :return: The text in doc-format with added entities
        """
        entity_list = self.regexes
        spans = []

        for entity in entity_list:
            for match in re.finditer(entity.regex_pattern, doc.text):
                start, end = match.span()
                span = doc.char_span(start, end, label=entity.label)
                if type(span) is not type(None):
                    if not self.overlap(span, doc):
                        spans.append(span)

        doc._.ents_regex = spans
        return doc

    def overlap(self, span, doc):
        """
        :param span: Span of words
        :param doc: The text in doc-format
        :return: True if overlap
        """
        for entity in doc.ents:
            if (entity.start <= span.start <= entity.end) or (entity.start <=span.end <=entity.end):
                return True
        return False



